#!/usr/bin/env python3

# Copyright (C) 2016 Red Hat
#
# This file is part of relval.
#
# relval is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""fedmsg consumer to create release validation events."""


import logging
import sys

import fedfind.helpers
import fedmsg
import fedmsg.config
import mwclient.errors
import smtplib
import wikitcms.event
import wikitcms.wiki

from six.moves.urllib.error import (URLError, HTTPError)

logger = logging.getLogger(__name__)


def consume(msg, test):
    """Consume incoming message. 'test' indicates test mode."""
    # only run on completed composes
    status = msg.get('status')
    location = msg.get('location')
    compstr = msg.get('compose_id', location)
    if not 'FINISHED' in status and location:
        return 1

    if test:
        server = 'stg.fedoraproject.org'
    else:
        server = 'fedoraproject.org'
    site = wikitcms.wiki.Wiki(('https', server), '/w/')
    site.login()

    # get the validation event instance
    try:
        event = site.get_validation_event_url(location)
    except ValueError:
        logger.info("Could not determine validation event for compose %s",
                    compstr)
        return 1

    # check if we're testable
    if not event.has_bootiso:
        logger.debug("Compose %s is not testable!", compstr)
        return 1

    # at present we should only ever create events for exactly the
    # release after the current stable release; this prevents us
    # creating Rawhide events during Branched periods. Note we have
    # to handle the case of creating the *first* Rawhide nightly
    # after a stable release, so we cannot just require release to
    # equal wiki.current_compose['release']
    curr = fedfind.helpers.get_current_release()
    if not int(event.release) == curr+1:
        logger.debug("Compose %s is not for the next release!", compstr)
        return 1

    # Make sure the potential new event is newer than the current one
    # (this avoids creating nightly events after TCs/RCs start)
    currev = site.current_event
    if not event.sorttuple > currev.sorttuple:
        logger.debug("Compose %s is not 'newer' than current compose!", compstr)
        return 1

    # if we're proposing a nightly event for the same release, do the
    # date check
    if not isinstance(event, wikitcms.event.ComposeEvent) and not int(event.release) > int(currev.release):
        try:
            newdate = fedfind.helpers.date_check(event.compose, out='obj')
            currdate = fedfind.helpers.date_check(currev.compose, out='obj')
            delta = newdate - currdate
        except ValueError:
            logger.warning("Could not determine date of current or new event!")
            return 1
        if delta.days < 3:
            # we never create an event if it's been < 3 days.
            logger.debug("Less than three days since current nightly event.")
            return 1
        elif delta.days > 14:
            # we *always* create an event if it's been > 14 days.
            logger.debug("More than two weeks since last nightly event!")
            diff = ''
        else:
            # between 3 and 14 we check if important packages have changed.
            # we could do this a bit smarter using the metadata, but OTOH
            # the RPMs metadata is like 90MiB per compose...maybe look
            # into doing it with PDC.
            logger.debug("Comparing package versions...")
            packages = ['anaconda', 'python-blivet', 'pyparted',
                        'lorax', 'pungi', 'parted', 'pykickstart']
            try:
                currpacks = currev.get_package_versions(packages)
                newpacks = event.get_package_versions(packages)
            except (ValueError, URLError, HTTPError):
                logger.debug("Package version check failed!")
                return 1
            if not currpacks:
                logger.debug("Could not do package version check on current compose!")
                return 1
            if not newpacks:
                logger.debug("Could not do package version check on new compose!")
                return 1
            for key in list(currpacks.keys()):
                if not currpacks[key] or not newpacks[key]:
                    logger.debug("Could not find versions for all significant "
                                 "packages in both composes!")
                    return 1
            if currpacks == newpacks:
                logger.debug("No significant package changes since last nightly event.")
                return 1
            else:
                logger.info("Significant package changes since last nightly event!")
                # this means we're carrying on.
                diff = ''
                for key in list(currpacks.keys()):
                    if currpacks[key] != newpacks[key]:
                        diff += ("{0} - {1}: {2}, {3}: {4}\n".format(
                            key, currev.compose, currpacks[key],
                            event.compose, newpacks[key]))

    logger.info("Creating validation event %s", event.version)
    try:
        event.create(check=True)
    except mwclient.errors.APIError as err:
        logger.warning("Mediawiki error creating event!")
        logger.debug("Error: %s", err)
        return 1
    except ValueError:
        logger.warning("Existing page found for event! Aborting.")
        return 1

    # send the announcement mail. FIXME: for now the text is still
    # nightly-specific, change that when we use this for TCs/RCs.
    logger.debug("Sending announcement email")
    dest = "test-announce@lists.fedoraproject.org"
    urltmpl = "https://fedoraproject.org/wiki/{0}"
    summurl = urltmpl.format(event.summary_page.name.replace(' ', '_'))
    pageurls = [urltmpl.format(pag.name.replace(' ', '_'))
                for pag in event.valid_pages if 'Summary' not in pag.name]
    if diff:
        difftext = '\nNotable package version changes:\n{0}'.format(diff)
    else:
        difftext = ''
    tmpl = ("""From: rawhide@fedoraproject.org
To: {0}
Subject: Fedora {1} nightly compose nominated for testing

Announcing the creation of a new nightly release validation test event
for Fedora {1}. Please help run some tests for this
nightly compose if you have time. For more information on nightly
release validation testing, see:
https://fedoraproject.org/wiki/QA:Release_validation_test_plan
{2}
Test coverage information for the current release can be seen at:
https://www.happyassassin.net/testcase_stats/{3}

You can see all results, find testing instructions and image download
locations, and enter results on the Summary page:

{4}

The individual test result pages are:

{5}

Thank you for testing!
-- 
Mail generated by relval: https://www.happyassassin.net/relval/
""")
    msg = tmpl.format(
        dest, event.version, difftext, event.release, summurl, '\n'.join(pageurls))
    if test:
        print(msg)
    else:
        server = smtplib.SMTP('localhost')
        server.sendmail('rawhide@fedoraproject.org', [dest], msg)


def main():
    """Main listener loop."""
    test = 'test' in sys.argv[1]
    if test:
        level = logging.DEBUG
        # for using fedmsg-dg-replay
        searchtop = "org.fedoraproject.dev.pungi.compose.status.change"
    else:
        level = logging.INFO
        searchtop = "org.fedoraproject.prod.pungi.compose.status.change"
    logging.basicConfig(level=level)
    config = fedmsg.config.load_config()
    try:
        for (_, _, topic, msg) in fedmsg.tail_messages(mute=True, **config):
            if topic == searchtop:
                consume(msg['msg'], test)
    except KeyboardInterrupt:
        sys.stderr.write("Interrupted, exiting...\n")
        sys.exit(1)


if __name__ == '__main__':
    main()
